@@ -14,6 +14,7 @@
 #include "../mptrack/TrackerSettings.h"
 #endif //MODPLUG_TRACKER
 #ifndef MODPLUG_NO_FILESAVE
+#include "mpt/io_file/fstream.hpp"
 #include "../common/mptFileIO.h"
 #endif
 #include "../common/misc_util.h"
@@ -30,8 +31,13 @@
 #include "mpt/io/base.hpp"
 #include "mpt/io/io.hpp"
 #include "mpt/io/io_stdstream.hpp"
+#include "mpt/parse/parse.hpp"
 //#include "mpt/crc/crc.hpp"
 #include "OggStream.h"
+#include <algorithm>
+#if MPT_PLATFORM_MULTITHREADED && !defined(MPT_COMPILER_QUIRK_NO_STDCPP_THREADS)
+#include <thread>
+#endif
 #ifdef MPT_WITH_OGG
 #if MPT_COMPILER_CLANG
 #pragma clang diagnostic push
@@ -66,19 +72,19 @@ OPENMPT_NAMESPACE_BEGIN
 
 struct FLACDecoder
 {
-	FileReader &file;
-	CSoundFile &sndFile;
-	SAMPLEINDEX sample;
-	bool ready;
+	FileReader &m_file;
+	CSoundFile &m_sndFile;
+	const SAMPLEINDEX m_sample;
+	bool m_ready = false;
 
-	FLACDecoder(FileReader &f, CSoundFile &sf, SAMPLEINDEX smp) : file(f), sndFile(sf), sample(smp), ready(false) { }
+	FLACDecoder(FileReader &f, CSoundFile &sf, SAMPLEINDEX smp) : m_file(f), m_sndFile(sf), m_sample(smp) { }
 
 	static FLAC__StreamDecoderReadStatus read_cb(const FLAC__StreamDecoder *, FLAC__byte buffer[], size_t *bytes, void *client_data)
 	{
-		FileReader &file = static_cast<FLACDecoder *>(client_data)->file;
+		FileReader &file = static_cast<FLACDecoder *>(client_data)->m_file;
 		if(*bytes > 0)
 		{
-			FileReader::off_t readBytes = *bytes;
+			FileReader::pos_type readBytes = *bytes;
 			LimitMax(readBytes, file.BytesLeft());
 			file.ReadRaw(mpt::byte_cast<mpt::byte_span>(mpt::span(buffer, readBytes)));
 			*bytes = readBytes;
@@ -94,8 +100,8 @@ struct FLACDecoder
 
 	static FLAC__StreamDecoderSeekStatus seek_cb(const FLAC__StreamDecoder *, FLAC__uint64 absolute_byte_offset, void *client_data)
 	{
-		FileReader &file = static_cast<FLACDecoder *>(client_data)->file;
-		if(!file.Seek(static_cast<FileReader::off_t>(absolute_byte_offset)))
+		FileReader &file = static_cast<FLACDecoder *>(client_data)->m_file;
+		if(!file.Seek(static_cast<FileReader::pos_type>(absolute_byte_offset)))
 			return FLAC__STREAM_DECODER_SEEK_STATUS_ERROR;
 		else
 			return FLAC__STREAM_DECODER_SEEK_STATUS_OK;
@@ -103,30 +109,30 @@ struct FLACDecoder
 
 	static FLAC__StreamDecoderTellStatus tell_cb(const FLAC__StreamDecoder *, FLAC__uint64 *absolute_byte_offset, void *client_data)
 	{
-		FileReader &file = static_cast<FLACDecoder *>(client_data)->file;
+		FileReader &file = static_cast<FLACDecoder *>(client_data)->m_file;
 		*absolute_byte_offset = file.GetPosition();
 		return FLAC__STREAM_DECODER_TELL_STATUS_OK;
 	}
 
 	static FLAC__StreamDecoderLengthStatus length_cb(const FLAC__StreamDecoder *, FLAC__uint64 *stream_length, void *client_data)
 	{
-		FileReader &file = static_cast<FLACDecoder *>(client_data)->file;
+		FileReader &file = static_cast<FLACDecoder *>(client_data)->m_file;
 		*stream_length = file.GetLength();
 		return FLAC__STREAM_DECODER_LENGTH_STATUS_OK;
 	}
 
 	static FLAC__bool eof_cb(const FLAC__StreamDecoder *, void *client_data)
 	{
-		FileReader &file = static_cast<FLACDecoder *>(client_data)->file;
+		FileReader &file = static_cast<FLACDecoder *>(client_data)->m_file;
 		return file.NoBytesLeft();
 	}
 
 	static FLAC__StreamDecoderWriteStatus write_cb(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 *const buffer[], void *client_data)
 	{
 		FLACDecoder &client = *static_cast<FLACDecoder *>(client_data);
-		ModSample &sample = client.sndFile.GetSample(client.sample);
+		ModSample &sample = client.m_sndFile.GetSample(client.m_sample);
 
-		if(frame->header.number.sample_number >= sample.nLength || !client.ready)
+		if(frame->header.number.sample_number >= sample.nLength || !client.m_ready)
 		{
 			// We're reading beyond the sample size already, or we aren't even ready to decode yet!
 			return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
@@ -176,25 +182,25 @@ struct FLACDecoder
 	static void metadata_cb(const FLAC__StreamDecoder *, const FLAC__StreamMetadata *metadata, void *client_data)
 	{
 		FLACDecoder &client = *static_cast<FLACDecoder *>(client_data);
-		if(client.sample > client.sndFile.GetNumSamples())
+		if(client.m_sample > client.m_sndFile.GetNumSamples())
 		{
-			client.sndFile.m_nSamples = client.sample;
+			client.m_sndFile.m_nSamples = client.m_sample;
 		}
-		ModSample &sample = client.sndFile.GetSample(client.sample);
+		ModSample &sample = client.m_sndFile.GetSample(client.m_sample);
 
 		if(metadata->type == FLAC__METADATA_TYPE_STREAMINFO && metadata->data.stream_info.total_samples != 0)
 		{
 			// Init sample information
-			client.sndFile.DestroySampleThreadsafe(client.sample);
-			client.sndFile.m_szNames[client.sample] = "";
+			client.m_sndFile.DestroySampleThreadsafe(client.m_sample);
+			client.m_sndFile.m_szNames[client.m_sample] = "";
 			sample.Initialize();
 			sample.uFlags.set(CHN_16BIT, metadata->data.stream_info.bits_per_sample > 8);
 			sample.uFlags.set(CHN_STEREO, metadata->data.stream_info.channels > 1);
 			sample.nLength = mpt::saturate_cast<SmpLength>(metadata->data.stream_info.total_samples);
 			LimitMax(sample.nLength, MAX_SAMPLE_LENGTH);
 			sample.nC5Speed = metadata->data.stream_info.sample_rate;
-			client.ready = (sample.AllocateSample() != 0);
-		} else if(metadata->type == FLAC__METADATA_TYPE_APPLICATION && !memcmp(metadata->data.application.id, "riff", 4) && client.ready)
+			client.m_ready = (sample.AllocateSample() != 0);
+		} else if(metadata->type == FLAC__METADATA_TYPE_APPLICATION && !memcmp(metadata->data.application.id, "riff", 4) && client.m_ready)
 		{
 			// Try reading RIFF loop points and other sample information
 			FileReader data(mpt::as_span(metadata->data.application.data, metadata->length));
@@ -203,8 +209,8 @@ struct FLACDecoder
 			// We're not really going to read a WAV file here because there will be only one RIFF chunk per metadata event, but we can still re-use the code for parsing RIFF metadata...
 			WAVReader riffReader(data);
 			riffReader.FindMetadataChunks(chunks);
-			riffReader.ApplySampleSettings(sample, client.sndFile.GetCharsetInternal(), client.sndFile.m_szNames[client.sample]);
-		} else if(metadata->type == FLAC__METADATA_TYPE_VORBIS_COMMENT && client.ready)
+			riffReader.ApplySampleSettings(sample, client.m_sndFile.GetCharsetInternal(), client.m_sndFile.m_szNames[client.m_sample]);
+		} else if(metadata->type == FLAC__METADATA_TYPE_VORBIS_COMMENT && client.m_ready)
 		{
 			// Try reading Vorbis Comments for sample title, sample rate and loop points
 			SmpLength loopStart = 0, loopLength = 0;
@@ -214,17 +220,17 @@ struct FLACDecoder
 				const FLAC__uint32 length = metadata->data.vorbis_comment.comments[i].length;
 				if(length > 6 && !mpt::CompareNoCaseAscii(tag, "TITLE=", 6))
 				{
-					client.sndFile.m_szNames[client.sample] = mpt::ToCharset(client.sndFile.GetCharsetInternal(), mpt::Charset::UTF8, mpt::String::ReadBuf(mpt::String::maybeNullTerminated, tag + 6, length - 6));
+					client.m_sndFile.m_szNames[client.m_sample] = mpt::ToCharset(client.m_sndFile.GetCharsetInternal(), mpt::Charset::UTF8, mpt::String::ReadBuf(mpt::String::maybeNullTerminated, tag + 6, length - 6));
 				} else if(length > 11 && !mpt::CompareNoCaseAscii(tag, "SAMPLERATE=", 11))
 				{
-					uint32 sampleRate = ConvertStrTo<uint32>(tag + 11);
+					uint32 sampleRate = mpt::parse<uint32>(tag + 11);
 					if(sampleRate > 0) sample.nC5Speed = sampleRate;
 				} else if(length > 10 && !mpt::CompareNoCaseAscii(tag, "LOOPSTART=", 10))
 				{
-					loopStart = ConvertStrTo<SmpLength>(tag + 10);
+					loopStart = mpt::parse<SmpLength>(tag + 10);
 				} else if(length > 11 && !mpt::CompareNoCaseAscii(tag, "LOOPLENGTH=", 11))
 				{
-					loopLength = ConvertStrTo<SmpLength>(tag + 11);
+					loopLength = mpt::parse<SmpLength>(tag + 11);
 				}
 			}
 			if(loopLength > 0)
@@ -458,7 +464,7 @@ bool CSoundFile::ReadFLACSample(SAMPLEINDEX sample, FileReader &file)
 	FLAC__stream_decoder_finish(decoder);
 	FLAC__stream_decoder_delete(decoder);
 
-	if(client.ready && Samples[sample].HasSampleData())
+	if(client.m_ready && Samples[sample].HasSampleData())
 	{
 		Samples[sample].Convert(MOD_TYPE_IT, GetType());
 		Samples[sample].PrecomputeLoops(*this, false);
@@ -490,7 +496,7 @@ struct FLAC__StreamEncoder_RAII
 
 	static FLAC__StreamEncoderWriteStatus StreamEncoderWriteCallback(const FLAC__StreamEncoder *encoder, const FLAC__byte buffer[], size_t bytes, unsigned samples, unsigned current_frame, void *client_data)
 	{
-		mpt::ofstream & file = *reinterpret_cast<mpt::ofstream*>(client_data);
+		mpt::IO::ofstream & file = *mpt::void_ptr<mpt::IO::ofstream>(client_data);
 		MPT_UNUSED_VARIABLE(encoder);
 		MPT_UNUSED_VARIABLE(samples);
 		MPT_UNUSED_VARIABLE(current_frame);
@@ -502,7 +508,7 @@ struct FLAC__StreamEncoder_RAII
 	}
 	static FLAC__StreamEncoderSeekStatus StreamEncoderSeekCallback(const FLAC__StreamEncoder *encoder, FLAC__uint64 absolute_byte_offset, void *client_data)
 	{
-		mpt::ofstream & file = *reinterpret_cast<mpt::ofstream*>(client_data);
+		mpt::IO::ofstream & file = *mpt::void_ptr<mpt::IO::ofstream>(client_data);
 		MPT_UNUSED_VARIABLE(encoder);
 		if(!mpt::in_range<mpt::IO::Offset>(absolute_byte_offset))
 		{
@@ -516,7 +522,7 @@ struct FLAC__StreamEncoder_RAII
 	}
 	static FLAC__StreamEncoderTellStatus StreamEncoderTellCallback(const FLAC__StreamEncoder *encoder, FLAC__uint64 *absolute_byte_offset, void *client_data)
 	{
-		mpt::ofstream & file = *reinterpret_cast<mpt::ofstream*>(client_data);
+		mpt::IO::ofstream & file = *mpt::void_ptr<mpt::IO::ofstream>(client_data);
 		MPT_UNUSED_VARIABLE(encoder);
 		mpt::IO::Offset pos = mpt::IO::TellWrite(file);
 		if(pos < 0)
@@ -557,7 +563,7 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 {
 #ifdef MPT_WITH_FLAC
 	const ModSample &sample = Samples[nSample];
-	if(sample.uFlags[CHN_ADLIB])
+	if(sample.uFlags[CHN_ADLIB] || !sample.HasSampleData())
 		return false;
 
 	FLAC__StreamEncoder_RAII encoder(f);
@@ -586,7 +592,7 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 		FLAC__metadata_object_vorbiscomment_append_comment(metadata[0], entry, false);
 		if(sampleRate > FLAC__MAX_SAMPLE_RATE)
 		{
-			// FLAC only supports a sample rate of up to 655350 Hz.
+			// FLAC only supports a sample rate of up to 1048575 Hz.
 			// Store the real sample rate in a custom Vorbis comment.
 			FLAC__metadata_object_vorbiscomment_entry_from_name_value_pair(&entry, "SAMPLERATE", mpt::afmt::val(sampleRate).c_str());
 			FLAC__metadata_object_vorbiscomment_append_comment(metadata[0], entry, false);
@@ -638,6 +644,11 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 		{
 			chunk.loops[chunk.info.numLoops++].ConvertToWAV(sample.nLoopStart, sample.nLoopEnd, sample.uFlags[CHN_PINGPONGLOOP]);
 			chunk.header.length += sizeof(WAVSampleLoop);
+		} else if(sample.uFlags[CHN_SUSTAINLOOP])
+		{
+			// Invent zero-length loop to distinguish sustain loop from normal loop
+			chunk.loops[chunk.info.numLoops++].ConvertToWAV(0, 0, false);
+			chunk.header.length += sizeof(WAVSampleLoop);
 		}
 
 		const uint32 length = sizeof(RIFFChunk) + chunk.header.length;
@@ -663,7 +674,7 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 
 		for(uint32 i = 0; i < std::size(sample.cues); i++)
 		{
-			chunk.cues[i].ConvertToWAV(i, sample.cues[i]);
+			chunk.cues[i] = ConvertToWAVCuePoint(i, sample.cues[i]);
 		}
 
 		const uint32 length = sizeof(RIFFChunk) + chunk.header.length;
@@ -672,12 +683,12 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 		numBlocks++;
 	}
 
-	// FLAC allows a maximum sample rate of 655350 Hz.
+	// FLAC allows a maximum sample rate of 1048575 Hz.
 	// If the real rate is higher, we store it in a Vorbis comment above.
 	LimitMax(sampleRate, FLAC__MAX_SAMPLE_RATE);
 	if(!FLAC__format_sample_rate_is_subset(sampleRate))
 	{
-		// FLAC only supports 10 Hz granularity for frequencies above 65535 Hz if the streamable subset is chosen.
+		// FLAC only supports 10 Hz granularity for frequencies above 65535 Hz if the streamable subset is chosen, and only a maximum frequency of 655350 Hz.
 		FLAC__stream_encoder_set_streamable_subset(encoder, false);
 	}
 	FLAC__stream_encoder_set_channels(encoder, sample.GetNumChannels());
@@ -687,6 +698,15 @@ bool CSoundFile::SaveFLACSample(SAMPLEINDEX nSample, std::ostream &f) const
 	FLAC__stream_encoder_set_metadata(encoder, metadata.data(), numBlocks);
 #ifdef MODPLUG_TRACKER
 	FLAC__stream_encoder_set_compression_level(encoder, TrackerSettings::Instance().m_FLACCompressionLevel);
+#if (FLAC_API_VERSION_CURRENT >= 14) && MPT_PLATFORM_MULTITHREADED && !defined(MPT_COMPILER_QUIRK_NO_STDCPP_THREADS)
+	uint32 threads = TrackerSettings::Instance().m_FLACMultithreading ? static_cast<uint32>(std::max(std::thread::hardware_concurrency(), static_cast<unsigned int>(1))) : static_cast<uint32>(1);
+	// Work-around <https://github.com/xiph/flac/issues/823>.
+	//FLAC__stream_encoder_set_num_threads(encoder, threads);
+	while((FLAC__stream_encoder_set_num_threads(encoder, threads) == FLAC__STREAM_ENCODER_SET_NUM_THREADS_TOO_MANY_THREADS) && (threads > 1))
+	{
+		threads = ((threads > 256) ? 256 : (threads - 1));
+	}
+#endif
 #endif // MODPLUG_TRACKER
 
 	bool success = FLAC__stream_encoder_init_stream(encoder, &FLAC__StreamEncoder_RAII::StreamEncoderWriteCallback, &FLAC__StreamEncoder_RAII::StreamEncoderSeekCallback, &FLAC__StreamEncoder_RAII::StreamEncoderTellCallback, nullptr, &encoder.f) == FLAC__STREAM_ENCODER_INIT_STATUS_OK;
